import unittest
import numpy as np
from numpy.testing import assert_array_equal
from EIMTC.preprocessing import OneHotEncoderEIMTC, FilenameBasedLabelling, DirectoryBasedLabelling


class TestFilenameBasedLabelling(unittest.TestCase):

    def test_filename_based_labelling1(self):
        # Given
        filepath = r'D:/dir1/dir2/filename1.pcap'
        label_name = 'lbl_name1'
        method = FilenameBasedLabelling(label_name)
        # When
        res = method(filepath)
        # Then
        self.assertEqual(res, {'lbl_name1': 'filename1'})
        
    def test_filename_based_labelling2(self):
        # Given
        filepath = r'D:/dir1/dir2/somethingNAME.pcap'
        label_name = 'lbl_name1'
        method = FilenameBasedLabelling(label_name)
        # When
        res = method(filepath)
        # Then
        self.assertEqual(res, {'lbl_name1': 'somethingNAME'})



class TestDirectoryBasedLabelling(unittest.TestCase):

    def test_directory_based_labelling_single(self):
        # Given
        filepath = r'D:/dir1/lbldir1/filename1.pcap'
        root = r'D:/dir1/'
        label_names = ['lbl_name1']
        method = DirectoryBasedLabelling(root, label_names)
        # When
        res = method(filepath)
        # Then
        self.assertEqual(res, {'lbl_name1': 'lbldir1'})
        
    def test_directory_based_labelling_2_layers_deep(self):
        # Given
        filepath = r'D:/dir1/lbldir1/lbldir2/filename1.pcap'
        root = r'D:/dir1/'
        label_names = ['lbl_name1', 'lbl_name2']
        method = DirectoryBasedLabelling(root, label_names)
        # When
        res = method(filepath)
        # Then
        self.assertEqual(res, {
            'lbl_name1': 'lbldir1',
            'lbl_name2': 'lbldir2'
        })



class TestOneHotEncoderEIMTC(unittest.TestCase):

    def test_onehotencoder_categories(self):
        # Given
        enc = OneHotEncoderEIMTC(sparse=False)
        X = sorted(['Male', 'Female', 'Other'])
        enc.fit(X)
        # When
        res = enc.categories_
        # Then
        assert_array_equal(res, X)
        
    def test_onehotencoder_fit_transform(self):
        # Given
        enc = OneHotEncoderEIMTC(sparse=False)
        X = sorted(['Male', 'Female', 'Other'])
        # When
        res = enc.fit_transform(X)
        # Then
        assert_array_equal(res, np.array([
            [1,0,0],
            [0,1,0],
            [0,0,1],
        ]))
        
    def test_onehotencoder_fit_transform_multiple_samples(self):
        # Given
        enc = OneHotEncoderEIMTC(sparse=False)
        X = ['Male', 'Male', 'Other', 'Female', 'Other']
        # When
        res = enc.fit_transform(X)
        # Then
        assert_array_equal(res, np.array([
            [0,1,0],
            [0,1,0],
            [0,0,1],
            [1,0,0],
            [0,0,1],
        ]))


    def test_onehotencoder_fit_and_then_transform_multiple_samples(self):
        # Given
        enc = OneHotEncoderEIMTC(sparse=False)
        X = ['Male', 'Male', 'Other', 'Female', 'Other']
        X2 = ['Male', 'Other', 'Male']
        enc.fit(X)
        # When
        res = enc.transform(X2)
        # Then
        assert_array_equal(res, np.array([
            [0,1,0],
            [0,0,1],
            [0,1,0],
        ]))
        
    def test_onehotencoder_inverse_transform(self):
        # Given
        enc = OneHotEncoderEIMTC(sparse=False)
        X = ['Male', 'Male', 'Other', 'Female', 'Other']
        transformed = [
            [0,1,0],
            [0,0,1],
            [0,1,0]
        ]
        enc.fit(X)
        # When
        res = enc.inverse_transform(transformed)
        # Then
        assert_array_equal(res, ['Male', 'Other', 'Male'])


import unittest
from os import path
import pandas as pd
import numpy as np
from numpy.testing import assert_array_equal
from EIMTC.tls_tshark_entry import _extract_tls_features, _read_config_file

dirname = path.dirname(__file__)
pcaps_dir = path.join(dirname, 'pcaps')


class Test3rdPartyTSharkForTLS(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.config = _read_config_file(
            path.join(path.dirname(__file__), '../tools/config.json')
        )
        
    def test_tls_client_hello_single_packet(self):
        # Given
        pcap_filepath = path.join(pcaps_dir, 'tls_client_hello_single_packet.pcap')
        # When
        df: pd.DataFrame = _extract_tls_features(pcap_filepath, self.config['tshark_location'])
        # Then
        self.assertFalse(df.empty) 
        ## Stats
        ## Stats | Bidirectional
        assert_array_equal(df['bidirectional_tls_records'], [1])
        assert_array_equal(df['bidirectional_tls_payload_bytes'], [512])
        assert_array_equal(df['bidirectional_tls_record_distinct_sizes'], [1])
        assert_array_equal(df['bidirectional_mean_tls_record_size'], [512])
        assert_array_equal(df['bidirectional_median_tls_record_size'], [512])
        assert_array_equal(df['bidirectional_stddev_tls_record_size'], [np.nan])
        assert_array_equal(df['bidirectional_variance_tls_record_size'], [np.nan])
        assert_array_equal(df['bidirectional_skew_from_median_tls_record_size'], [np.nan])
        assert_array_equal(df['bidirectional_coeff_of_var_tls_record_size'], [np.nan])
        assert_array_equal(df['bidirectional_min_tls_record_size'], [512])
        assert_array_equal(df['bidirectional_max_tls_record_size'], [512])
        ## Stats | src -> dst
        assert_array_equal(df['src2dst_tls_records'], [1])
        assert_array_equal(df['src2dst_tls_payload_bytes'], [512])
        assert_array_equal(df['src2dst_tls_record_distinct_sizes'], [1])
        assert_array_equal(df['src2dst_mean_tls_record_size'], [512])
        assert_array_equal(df['src2dst_median_tls_record_size'], [512])
        assert_array_equal(df['src2dst_stddev_tls_record_size'], [np.nan])
        assert_array_equal(df['src2dst_variance_tls_record_size'], [np.nan])
        assert_array_equal(df['src2dst_skew_from_median_tls_record_size'], [np.nan])
        assert_array_equal(df['src2dst_coeff_of_var_tls_record_size'], [np.nan])
        assert_array_equal(df['src2dst_min_tls_record_size'], [512])
        assert_array_equal(df['src2dst_max_tls_record_size'], [512])
        ## Stats | dst -> src
        assert_array_equal(df['dst2src_tls_records'], [0])
        assert_array_equal(df['dst2src_tls_payload_bytes'], [0])
        assert_array_equal(df['dst2src_tls_record_distinct_sizes'], [0])
        assert_array_equal(df['dst2src_mean_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_median_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_stddev_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_variance_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_skew_from_median_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_coeff_of_var_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_min_tls_record_size'], [np.nan])
        assert_array_equal(df['dst2src_max_tls_record_size'], [np.nan])
        ## Clumps
        ## Clumps | Bidirectional
        assert_array_equal(df['bidirectional_tls_clumps'], [1])
        assert_array_equal(df['bidirectional_mean_tls_clump_bytes'], [512])
        assert_array_equal(df['bidirectional_median_tls_clump_bytes'], [512])
        assert_array_equal(df['bidirectional_stddev_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['bidirectional_variance_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['bidirectional_skew_from_median_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['bidirectional_coeff_of_var_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['bidirectional_min_tls_clump_bytes'], [512])
        assert_array_equal(df['bidirectional_max_tls_clump_bytes'], [512])
        assert_array_equal(df['bidirectional_mean_tls_clump_sizes'], [1])
        assert_array_equal(df['bidirectional_median_tls_clump_sizes'], [1])
        assert_array_equal(df['bidirectional_stddev_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['bidirectional_variance_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['bidirectional_skew_from_median_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['bidirectional_coeff_of_var_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['bidirectional_min_tls_clump_sizes'], [1])
        assert_array_equal(df['bidirectional_max_tls_clump_sizes'], [1])
        ## Clumps | src -> dst
        assert_array_equal(df['src2dst_tls_clumps'], [1])
        assert_array_equal(df['src2dst_mean_tls_clump_bytes'], [512])
        assert_array_equal(df['src2dst_median_tls_clump_bytes'], [512])
        assert_array_equal(df['src2dst_stddev_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['src2dst_variance_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['src2dst_skew_from_median_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['src2dst_coeff_of_var_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['src2dst_min_tls_clump_bytes'], [512])
        assert_array_equal(df['src2dst_max_tls_clump_bytes'], [512])
        assert_array_equal(df['src2dst_mean_tls_clump_sizes'], [1])
        assert_array_equal(df['src2dst_median_tls_clump_sizes'], [1])
        assert_array_equal(df['src2dst_stddev_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['src2dst_variance_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['src2dst_skew_from_median_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['src2dst_coeff_of_var_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['src2dst_min_tls_clump_sizes'], [1])
        assert_array_equal(df['src2dst_max_tls_clump_sizes'], [1])
        ## Clumps | dst -> src
        assert_array_equal(df['dst2src_tls_clumps'], [0])
        assert_array_equal(df['dst2src_mean_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_median_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_stddev_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_variance_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_skew_from_median_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_coeff_of_var_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_min_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_max_tls_clump_bytes'], [np.nan])
        assert_array_equal(df['dst2src_mean_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_median_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_stddev_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_variance_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_skew_from_median_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_coeff_of_var_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_min_tls_clump_sizes'], [np.nan])
        assert_array_equal(df['dst2src_max_tls_clump_sizes'], [np.nan])

    def test_tls_record_per_packet_single_session(self):
        # Given
        pcap_filepath = path.join(pcaps_dir, 'tls_pkt_rel_time_single.pcap')
        # When
        df: pd.DataFrame = _extract_tls_features(pcap_filepath, self.config['tshark_location'])
        # Then
        self.assertFalse(df.empty) 
        self.assertSequenceEqual(list(df['bidirectional_tls_records']), [8])
        self.assertSequenceEqual(list(df['bidirectional_tls_payload_bytes']), [836])
        self.assertSequenceEqual(list(df['bidirectional_tls_record_distinct_sizes']), [2])
        self.assertSequenceEqual(list(df['bidirectional_mean_tls_record_size']), [104.5])
        self.assertSequenceEqual(list(df['bidirectional_min_tls_record_size']), [39])
        self.assertSequenceEqual(list(df['bidirectional_max_tls_record_size']), [170])
    
    def test_tls_multirecord_in_packet_single_session(self):
        # Given
        pcap_filepath = path.join(pcaps_dir, 'tls_small_pkt_payload_ratio_single.pcap')
        # When
        df: pd.DataFrame = _extract_tls_features(pcap_filepath, self.config['tshark_location'])
        # Then
        self.assertFalse(df.empty) 
        self.assertSequenceEqual(list(df['bidirectional_tls_records']), [12])
        self.assertSequenceEqual(list(df['bidirectional_tls_payload_bytes']), [1809])
        self.assertSequenceEqual(list(df['bidirectional_tls_record_distinct_sizes']), [10])
        self.assertSequenceEqual(list(df['bidirectional_mean_tls_record_size']), [150.75])
        self.assertSequenceEqual(list(df['bidirectional_min_tls_record_size']), [1])
        self.assertSequenceEqual(list(df['bidirectional_max_tls_record_size']), [664])



from runstats import *
from nfstream import NFPlugin
import numpy as np

class Packets_size_and_interarrival_time(NFPlugin):
    ''' Packets_size_and_interarrival_time |
    Extracts statistics of packet size (raw) and inter-arrival time (ms).
    
    Feature outputs:
        Statistics of packet size and IAT of the first 'n_packets' packets of the flow.
        
        Name Format:
            udps.packets_[FEATURE]_[STAT]
        
        Statistics (STAT) Names:
            - max
            - min
            - mean
            - stddev
            - skewness
            - variance
            - kurtosis
            - sum
            - first_quartile
            - second_quartile
            - third_quartile
            
        Feature Names:
            - size
            - interarrival_time
    
    Paper: 
        "Deep Learning for Network Traffic Classification".
    
        By:
            - Niloofar Bayat.
            - Weston Jackson.
            - Derrick Liu.
            
            Columbia University.
    
    Contribution Requests:
        1. Optimize median calculations by using histogram (if possible)
        
    '''
    def __init__(self, n_packets=None):
        '''
        Args:
            `n_packets` (int): The number of packets to process. If None then the whole flow will be processed.
        '''
        self.n_packets = n_packets
    
    def on_init(self, packet, flow):
        ''' '''
        flow.udps.packets_size = list()
        flow.udps.packets_interarrival_time = list()
        self.on_update(packet,flow)

    def on_update(self, packet, flow):
        ''' '''
        if self.n_packets is not None and flow.bidirectional_packets > self.n_packets:
            return
        
        flow.udps.packets_size.append(packet.raw_size)
        flow.udps.packets_interarrival_time.append(packet.delta_time)

    def on_expire(self, flow):
        ''' '''
        packets_size_statistics = Statistics(flow.udps.packets_size)
        packets_interarrival_time_statistics = Statistics(flow.udps.packets_interarrival_time)

        #Packets Size Statistical
        packets_size_Q1,packets_size_Q2,packets_size_Q3 = quartile(np.array(flow.udps.packets_size))
        flow.udps.packets_size_min = packets_size_statistics.minimum() if packets_size_statistics._count > 0 else 0
        flow.udps.packets_size_max = packets_size_statistics.maximum() if packets_size_statistics._count > 0 else 0
        flow.udps.packets_size_stddev = packets_size_statistics.stddev() if packets_size_statistics._count >= 2 else 0
        flow.udps.packets_size_first_quartile = packets_size_Q1
        flow.udps.packets_size_second_quartile = packets_size_Q2
        flow.udps.packets_size_third_quartile = packets_size_Q3
        flow.udps.packets_size_mean = packets_size_statistics.mean()
        flow.udps.packets_size_median_absoulte_deviation = median_absolute_deviation(np.array(flow.udps.packets_size))
        flow.udps.packets_size_variance = packets_size_statistics.variance() if packets_size_statistics._count >= 2 else 0
        flow.udps.packets_size_skewness = packets_size_statistics.skewness() if flow.udps.packets_size_stddev != 0 else 0
        flow.udps.packets_size_kurtosis = packets_size_statistics.kurtosis() if flow.udps.packets_size_stddev != 0 else 0
        flow.udps.packets_size_sum = sum(flow.udps.packets_size)
        
        #Packets Interarrival Time Statistical
        packets_interarrival_time_Q1,packets_interarrival_time_Q2,packets_interarrival_time_Q3 = quartile(np.array(flow.udps.packets_interarrival_time))
        flow.udps.packets_interarrival_time_min = packets_interarrival_time_statistics.minimum() if packets_interarrival_time_statistics._count > 0 else 0
        flow.udps.packets_interarrival_time_max = packets_interarrival_time_statistics.maximum() if packets_interarrival_time_statistics._count > 0 else 0
        flow.udps.packets_interarrival_time_stddev = packets_interarrival_time_statistics.stddev() if packets_interarrival_time_statistics._count >= 2 else 0
        flow.udps.packets_interarrival_time_first_quartile = packets_interarrival_time_Q1
        flow.udps.packets_interarrival_time_second_quartile = packets_interarrival_time_Q2
        flow.udps.packets_interarrival_time_third_quartile = packets_interarrival_time_Q3
        flow.udps.packets_interarrival_time_mean = packets_interarrival_time_statistics.mean()
        flow.udps.packets_interarrival_time_median_absoulte_deviation = median_absolute_deviation(np.array(flow.udps.packets_interarrival_time))
        flow.udps.packets_interarrival_time_variance = packets_interarrival_time_statistics.variance() if packets_interarrival_time_statistics._count >= 2 else 0
        flow.udps.packets_interarrival_time_skewness = packets_interarrival_time_statistics.skewness() if flow.udps.packets_interarrival_time_stddev != 0 else 0
        flow.udps.packets_interarrival_time_kurtosis = packets_interarrival_time_statistics.kurtosis() if flow.udps.packets_interarrival_time_stddev != 0 else 0
        flow.udps.packets_interarrival_time_sum = sum(flow.udps.packets_interarrival_time)

def median_absolute_deviation(arr):
    return np.median(np.absolute(arr-np.median(arr)))

def quartile(arr):
    sorted_arr = np.sort(arr)
    median = sorted_median(sorted_arr)
    sorted_len = len(sorted_arr)
    half = int(sorted_len/2)
    if sorted_len % 2 == 0:
        return sorted_median(sorted_arr[0:half]),median,sorted_median(sorted_arr[half:sorted_len])
    else:
        return sorted_median(sorted_arr[0:half+1]),median,sorted_median(sorted_arr[half:sorted_len])

def sorted_median(sorted_arr):
    arr_len = len(sorted_arr)
    half = int(arr_len/2)
    if arr_len % 2 == 0:
        return (sorted_arr[half-1]+sorted_arr[half])/2
    else:
        return sorted_arr[half]


from ._m1cnn import M1CNNPreprocessing
from ._distiller import DistillerPreprocessing
from ._maldist import MalDistPreprocessing
# wrappers
from ._wrappers import OneHotEncoderEIMTC
# labellings
from ._labelling import (
    FilenameBasedLabelling, DirectoryBasedLabelling
)
# general purpose 
from sklearn.preprocessing import (
    add_dummy_feature, binarize, Binarizer, FunctionTransformer, KBinsDiscretizer, 
    LabelBinarizer, label_binarize, KernelCenterer, LabelEncoder, maxabs_scale, 
    MaxAbsScaler, minmax_scale, MinMaxScaler, MultiLabelBinarizer, normalize, 
    Normalizer, OneHotEncoder, OrdinalEncoder, StandardScaler
)

__all__ = [
    'M1CNNPreprocessing',
    'DistillerPreprocessing',
    'MalDistPreprocessing'
] + [ # wrappers
    'OneHotEncoderEIMTC'
] + [ # labellings
    'FilenameBasedLabelling', 
    'DirectoryBasedLabelling'
] + [ # general purpose
    'add_dummy_feature',
    'binarize, Binarizer',
    'FunctionTransformer',
    'KBinsDiscretizer, LabelBinarizer',
    'label_binarize',
    'KernelCenterer',
    'LabelEncoder',
    'maxabs_scale',
    'MaxAbsScaler',
    'minmax_scale',
    'MinMaxScaler',
    'MultiLabelBinarizer',
    'normalize',
    'Normalizer',
    'OneHotEncoder',
    'OrdinalEncoder',
    'StandardScaler'
]




import pandas as pd
import numpy as np
import ast
from .tls_stats import TLSRecordClumpStats, TLSRecordStats


def merge_df_by_biflows(df_tshark, df_nfstream):
    def add_five_tuple_biflow_id_column_nfstream(df):
        df['five_tuple_biflow_id'] = df.apply(lambda r: sort_key_string(r[['src_ip', 'src_port', 'dst_ip', 'dst_port', 'protocol']]), axis=1)
        return df
    
    df_nfstream = add_five_tuple_biflow_id_column_nfstream(df_nfstream)
    df_tshark = df_tshark.reset_index()
    df_merged = df_nfstream.merge(df_tshark, left_on='five_tuple_biflow_id', right_on='five_tuple_biflow_id', how='left')
    df_merged.drop(columns='five_tuple_biflow_id', inplace=True)
    return df_merged


def read_tshark_csv_output(filepath):
    df = pd.read_csv(filepath, dtype={'tls.record.length': str})
    return df


def get_tls_record_stats(df):
    grouped = groupby_biflows(df)
    return grouped.apply(
        lambda c: TLSRecordStats(
            c[['tls.record.length','direction']].explode('tls.record.length')
        ).length_stats()
    )


def get_tls_clump_stats(df):
    grouped = groupby_biflows(df)
    return grouped.apply(
        lambda g: TLSRecordClumpStats(
            g[['tls.record.length','direction', 'clump_num']].explode('tls.record.length')
        ).clump_stats()
    )
    

def drop_lengthless_tls_records(df):
    '''can happen with tcp restrasnsmission and network errors'''
    return df.dropna(subset=['tls.record.length'])


def convert_to_correct_dtypes(df):
    '''
    tls.record.length can be scalar/list, but it is represented as string at load.
    '''
    if df['tls.record.length'].dtype != 'O':
        # if it contains only scalars, pandas would recognize is the dtype
        # as int64 instead of str/object.
        df['tls.record.length'] = df['tls.record.length'].apply(str)
    df['tls.record.length'] = df['tls.record.length'].apply(ast.literal_eval)
    df['tls.record.length']
    return df


def add_packet_directions_column(df):
    def get_packet_direction(group):
        first_packet = group.iloc[0]
        (src, dst) = first_packet[['ip.src', 'ip.dst']]
        src2dst_packets_indices = group[group['ip.src'] == src].index
        src2dst_packets_direction = np.array([np.zeros(len(src2dst_packets_indices)), src2dst_packets_indices])
        dst2src_packets_indices = group[group['ip.src'] == dst].index
        dst2src_packets_direction = np.array([np.full(len(dst2src_packets_indices), 1), dst2src_packets_indices])
        return pd.Series(
            np.concatenate([src2dst_packets_direction[0], dst2src_packets_direction[0]]), 
            index=np.concatenate([src2dst_packets_direction[1], dst2src_packets_direction[1]])
        )
        
    grouped = groupby_biflows(df)
    directions = grouped.apply(get_packet_direction)
    if directions.index.nlevels == 1:
        ''' if there is only a single session, then pandas
        converts the packet number to columns '''
        directions = directions.iloc[0]
    else: 
        ''' There should be a multindex of fiveuple id string
        along with he packet number, so we remove the fivuple.
        '''
        directions = directions.droplevel(0)
    df['direction'] = directions
    

    return df


def add_packet_clump_num_column(df):
    def get_clump_nums(session):
        clump_nums = []
        df_indices = []
        dir = 0
        begin = 0
        end = 0
        clump_num = 0
        for i, packet in enumerate(session.itertuples()):
            if dir != packet.direction:
                end = i
                clump_nums.extend([clump_num]*(end-begin))
                df_indices.extend(session.index[begin:end])
                begin = end
                dir = packet.direction
                clump_num += 1
        if end <= begin:
            clump_nums.extend([clump_num]*(len(session)-begin))
            df_indices.extend(session.index[begin:])
        return pd.Series(clump_nums, dtype=object, index=df_indices)
    
    grouped = groupby_biflows(df)
    clump_nums = grouped.apply(get_clump_nums)
    if clump_nums.index.nlevels == 1:
        ''' if there is only a single session, then pandas
        converts the packet number to columns '''
        clump_nums = clump_nums.iloc[0]
    else: 
        ''' There should be a multindex of fiveuple id string
        along with he packet number, so we remove the fivuple.
        '''
        clump_nums = clump_nums.droplevel(0)
    df['clump_num'] = clump_nums

    return df


def groupby_biflows(df):
    '''
    TODO: add caching mechanism - groupby is a view of the dataframe.
    '''
    if not 'five_tuple_biflow_id' in df:
        df = add_five_tuple_biflow_id_column(df)
    return df.groupby('five_tuple_biflow_id', sort=False)


def add_five_tuple_biflow_id_column(df):
    df['five_tuple_biflow_id'] = df.apply(lambda r: sort_key_string(r[['ip.src', 'tcp.srcport', 'ip.dst', 'tcp.dstport', 'ip.proto']]), axis=1)
    return df


def sort_key_string(key):
    return '-.'.join(sorted(map(str, key)))






import numpy as np
from tensorflow.keras import activations
import tensorflow as tf
import tensorflow.keras as tfk


class ChainedGNN(tfk.Model):
    '''
    ## Paper
    
    "CGNN: Traffic Classification with Graph Neural Network"
    ### By
    - Bo Pang.
    - Yongquan Fu.
    - Siyuan Ren.
    - Ye Wang.
    - Qing Liao.
    - Yan Jia.
    '''
    def __init__(self, n_classes, use_precomputed_norm_adj_mat=True):
        super(ChainedGNN, self).__init__()
        self.use_precomputed_norm_adj_mat = use_precomputed_norm_adj_mat
        self.sgc1 = SGCLayer(516, 1, 'relu', use_precomputed_norm_adj_mat=use_precomputed_norm_adj_mat)
        self.sgc2 = SGCLayer(256, 1, 'relu', use_precomputed_norm_adj_mat=use_precomputed_norm_adj_mat)
        self.dense = tfk.layers.Dense(n_classes)

    def call(self, inputs):
        '''
        x = [A, F] or F. Depends on use_precomputed_norm_adj_mat
        '''
        x = self.sgc1(inputs)
        x = self.sgc2(x)
        if self.use_precomputed_norm_adj_mat:
            F = tf.expand_dims(x, -1)
        else:
            F = tf.expand_dims(x[1], -1)
        x = tfk.layers.AveragePooling2D(strides=1, pool_size=(F.shape[1],1))(F)
        x = tf.squeeze(x, [1, -1])
        x = self.dense(x)
        x = tf.nn.softmax(x)
        return x


class SGCLayer(tfk.layers.Layer):
    '''
    Simple Graph Convolution (SGC) Layer
    '''
    def __init__(self, outputsNumber, neighbourhood_distance=1, activation=None, use_precomputed_norm_adj_mat=True):
        super(SGCLayer, self).__init__()
        self.outputs = outputsNumber
        self.S = None # local normalized adj matrix
        self.neighbourhood_distance = neighbourhood_distance
        self.activation = activations.get(activation)
        self.use_precomputed_norm_adj_mat = use_precomputed_norm_adj_mat

    def build(self, input_shape):
        if self.use_precomputed_norm_adj_mat:
            node_feature_length = input_shape[-1]
            self.S = tf.Variable(
                initial_value=np.linalg.matrix_power(
                    getS(input_shape[1]), 
                    self.neighbourhood_distance
                ), 
                trainable=False,
                name='Normalized Adj. Matrix'
            )
        else: 
            node_feature_length = input_shape[1][-1]
            
        self.teta = self.add_weight(
            "teta",
            shape=[node_feature_length, self.outputs],
            trainable=True,
            initializer="random_normal"
        )
            
    def call(self, inputs):
        if self.use_precomputed_norm_adj_mat:
            return self._call_without_adj(inputs)
        else:
            return self._call_with_adj(inputs)
        
    def _call_with_adj(self, inputs):
        A = inputs[0]
        S = norm_adj_matrix(inputs[0])
        F = inputs[1] # Node features matrix
        F = self._call_work_node_features(S, F)
        
        return [A, F]
    
    def _call_without_adj(self, inputs):
        S = self.S
        F = inputs # Node features matrix
        F = self._call_work_node_features(S, F)
        
        return F
        
    def _call_work_node_features(self, S, F):
        F = tf.matmul(S, F)
        F = tf.matmul(F, self.teta)
        if self.activation is not None:
            F = self.activation(F)
        
        return F


def getS(n):
    # By itay and Omer
    # source: https://github.com/itay74121/GNN/blob/master/CGNN/CGNN.py
    if n == 1:
        return np.array(0)
    if n == 2:
        return np.array([(0, 1),(1, 0)])
    S = np.zeros((n,n), dtype='float32')
    S[0,0], S[n - 1,n - 1] = 0.50000, 0.50000 # first and last vertices with self loop
    S[0,1], S[1,0], S[n - 2,n - 1], S[n - 1,n - 2] = 0.40825, 0.40825, 0.40825, 0.40825 # first and last vertices with other edges
    S[1,1], S[1,2], S[n - 2,n - 2], S[n - 2,n - 3] = 0.33333, 0.33333, 0.33333, 0.33333 # rest connected to first and last
    for i in range(2, len(S) - 2): # rest of vertices
        for j in range(i - 1, i + 2):
            S[i,j] = 0.33333
    return S


def norm_adj_matrix(adj_batch, add_self_loops=True):
    # source: https://github.com/stellargraph/stellargraph/blob/5ca1e59e91cb6ac470bf19ff3da39b3a1a68650e/stellargraph/core/utils.py
    # Add self loops.
    if add_self_loops:
        adj_batch = adj_batch + tf.linalg.diag(tf.ones(adj_batch.shape[1]) - tf.linalg.diag_part(adj_batch))

    # Normalization
    rowsum = tf.reduce_sum(adj_batch, 1)
    d_mat_inv_sqrt = tf.linalg.diag(tf.math.rsqrt(rowsum))
    adj_normalized = tf.matmul(tf.matmul(d_mat_inv_sqrt, adj_batch), d_mat_inv_sqrt)
    
    return adj_normalized




# from tls_record_joy import TLSRecordJoy

from nfstream import NFStreamer  # https://www.nfstream.org/docs/api
from os import path
from .utils import is_iterable
from .tls_tshark_entry import extract_tls_features_and_merge
#from .hosts_processor import HostsProcessor
from .plugins.clump_flows import Clump_Flow
from .plugins.packets_size_interarrival_time import Packets_size_and_interarrival_time
from .plugins.asn_info import ASNInfo
from .plugins.n_pkts_byte_freq import NPacketsByteFrequency
from .plugins.first_packet_payload import FirstPacketPayloadLen
from .plugins.most_freq_payload_len_ratio import MostFreqPayloadLenRatio
from .plugins.dns_counter import DNSCounter
from .plugins.small_pkt_payload_ratio import SmallPacketPayloadRatio
from .plugins.pkt_rel_time import PacketRelativeTime
from .plugins.res_req_diff_time import ResReqDiffTime
#from .plugins.graypic import GrayPic1
from .plugins.protocol_header_fields import ProtocolHeaderFields
from .plugins.n_bytes import NBytes
from .plugins.stnn import STNN
#from guppy import hpy; h=hpy()


class Extractor:
    def __init__(self, output_dirpath, bpf_filter_string=None, custom_plugin_package=None, TLS=True) -> None:
        self.output_dirpath = output_dirpath
        self.bpf_filter_string = bpf_filter_string
        self.custom_plugin_package = custom_plugin_package
        self.TLS = TLS


    def extract(self, input_pcap_filepath, append=False):
        if is_iterable(input_pcap_filepath):
            self.extract_many(input_pcap_filepath, append)
        else:
            self.extract_single(input_pcap_filepath, append)

        return path.join(self.output_dirpath,'out-sessions.csv')
    
    
    def extract_single(self, input_pcap_filepath, labelling_method=None, append=False):
        plugins = self._plugins() if self.custom_plugin_package is None else self.custom_plugin_package
        session_streamer = NFStreamer(
            source=input_pcap_filepath,
            decode_tunnels=True,
            bpf_filter=self.bpf_filter_string,
            promiscuous_mode=True,
            snapshot_length=1536,
            idle_timeout=999999999,
            active_timeout=999999999,
            accounting_mode=0,
            udps=plugins,
            n_dissections=20,
            statistical_analysis=True,
            splt_analysis=0,
            n_meters=0,
            performance_report=0
        )

        sessions_df = session_streamer.to_pandas(columns_to_anonymize=[])
        if self.TLS:
            config_filepath = path.join(path.dirname(__file__), 'tools/config.json')
            sessions_df = extract_tls_features_and_merge(sessions_df, input_pcap_filepath, config_filepath)
        
        if labelling_method is not None:
            labels = labelling_method(input_pcap_filepath)
            for name, label in labels.items():
                sessions_df[name] = label
            
        sessions_df['filepath'] = input_pcap_filepath
        sessions_df.to_csv(
            path.join(self.output_dirpath, 'out-sessions.csv'), 
            index=False, 
            mode='a' if append else 'w', 
            header=not append
        )


    def extract_many(self, input_pcap_filepaths, labelling_method=None, append=False):
        for i, filepath in enumerate(self.extract_iter(input_pcap_filepaths, labelling_method, append)):
            print(f'[{i+1}/{len(input_pcap_filepaths)}] Finished extraction from {filepath}')


    def extract_iter(self, input_pcap_filepaths, labelling_method=None, append=False):
        for filepath in input_pcap_filepaths:
            self.extract_single(filepath, labelling_method, append)
            yield filepath
            append = True
            
            
    def _plugins(self):
        return [
            ASNInfo(),
            DNSCounter(),
            FirstPacketPayloadLen(),
            MostFreqPayloadLenRatio(),
            NPacketsByteFrequency(n_first_packets=6),
            PacketRelativeTime(),
            SmallPacketPayloadRatio(),
            ResReqDiffTime(),
            Clump_Flow(),
            Packets_size_and_interarrival_time(),
            ProtocolHeaderFields(20),
            NBytes(),
            STNN(20)
        ]


from nfstream import NFPlugin
import numpy as np
from scapy.all import IP, IPv6, raw
from pypacker.layer3 import ip
from pypacker.layer3 import ip6
import pandas as pd
import ast



class NBytes(NFPlugin):
    ''' NBytes |
    Extracts the first n_bytes from the flow, the bytes are taken
    from the transport layer payload (L4). if the flow have less than n_bytes bytes,
    then the rest of the bytes are zero-valued (padding).
        
    Paper:
        "End-to-end Encrypted Traffic Classification with One-dimensional Convolution Neural Networks".

        By:
            - Wei Wang.
            - Ming Zhu.
            - Jinlin Wang.
            - Xuewen Zeng.
            - Zhongzhen Yang.
    '''
    def __init__(self, n=784):
        '''
        Args:
            `n` (int): The number of bytes to extract from the flow's payload.
        '''
        self.n = n
    
    def on_init(self, packet, flow):
        ''' '''
        flow.udps.n_bytes_value = self.n
        flow.udps.n_bytes = [] # np.zeros(self.n)
        flow.udps.n_bytes_counted = 0
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        ''' '''
        remaining_bytes = self.n - flow.udps.n_bytes_counted
        if remaining_bytes >= 0 and packet.protocol in [6, 17]: # TCP or UDP only.
            amount_to_copy = min(remaining_bytes, packet.payload_size)
            if amount_to_copy == 0:
                return

            copied_binary_payload = self.get_payload_as_binary_pypacker(packet)[:amount_to_copy]
            flow.udps.n_bytes.extend(
                copied_binary_payload
            )
            flow.udps.n_bytes_counted += len(copied_binary_payload)

    def on_expire(self, flow):
        ''' '''
        '''
        Normalize to [0,1]: 
        flow.udps.n_bytes /= 255
        Optional cleanup: 
        del flow.udps.n_bytes_counted
        '''
        # Padding if necessary.
        if flow.udps.n_bytes_counted < self.n:
            remaining_bytes = self.n - flow.udps.n_bytes_counted
            flow.udps.n_bytes.extend(
                np.full(remaining_bytes, 0)
            )

    def get_payload_as_binary(self, packet):
        return packet.ip_packet[-packet.payload_size:]
    
    def get_payload_as_binary_scapy(self, packet):
        '''
        Older versions of NFStream (lower than 6.1) has problems with 
        detection the correct payload size of ipv6 packets.
        This function is here as a fallback in case of problems in future
        versions.

        in 6.4.2 and 6.4.3 there is a bug with wrong transport_size, also
        NFStream does not detect ethernet padding.
        '''
        ip_version = packet.ip_version
        if ip_version == 4:
            scapy_packet = IP(packet.ip_packet)
        elif ip_version == 6:
            scapy_packet = IPv6(packet.ip_packet)

        return raw(scapy_packet.payload.payload)
    
    def get_payload_as_binary_pypacker(self, packet):
        '''
        Scapy has a bug with correctly dissecting SNMP payload.
        '''
        ip_version = packet.ip_version
        if ip_version == 4:
            pypacker_packet = ip.IP(packet.ip_packet)
        elif ip_version == 6:
            pypacker_packet = ip6.IP6(packet.ip_packet)

        return pypacker_packet.upper_layer.body_bytes

    @staticmethod
    def preprocess(dataframe: pd.DataFrame):
        ''' 
        Preprocessing method for the n_bytes features.
        Converting 'udps.n_bytes' column from str to list.
        '''
        # validate
        assert 'udps.n_bytes' in dataframe.columns, "Column 'udps.n_bytes' not found."
        assert isinstance(dataframe['udps.n_bytes'].iloc[0], str), "Values in column 'udps.n_bytes' are already processed."
        
        dataframe['udps.n_bytes'] = dataframe['udps.n_bytes'].apply(ast.literal_eval)
        
        



from pypacker import ppcap
from pypacker.layer12 import ethernet
from pypacker.layer3 import ip
from pypacker.layer4 import tcp



def main():
    filepath = '../temp/DoH-Firefox84-NextDNS-1-pcap-format.pcap'
    output_path = '../temp/output_sess_ex.pcap'
    
    preader = ppcap.Reader(filename=filepath)
    pwriter = ppcap.Writer(filename=output_path, linktype=ppcap.DLT_EN10MB)

    for ts, buf in preader:
        eth = ethernet.Ethernet(buf)

        if eth[ip.IP] is not None:
            if eth[tcp.TCP] is not None:
                '192.168.1.105', 61296, '213.227.181.51', 443, 'tcp'
                if (eth[ip.IP].src_s == '192.168.1.105' or eth[ip.IP].dst_s == '192.168.1.105'
                    ) and (eth[ip.IP].src_s == '213.227.181.51' or eth[ip.IP].dst_s == '213.227.181.51'
                    ) and (eth[tcp.TCP].sport == 61296 or eth[tcp.TCP].sport == 443
                    ) and (eth[tcp.TCP].dport == 61296 or eth[tcp.TCP].dport == 443):
                    
                #print("%d: %s:%s -> %s:" % (ts, eth[ip.IP].src_s, eth[ip.IP].dst_s, eth[tcp.TCP].sport))
                    pwriter.write(eth.bin())

    pwriter.close()



if __name__ == '__main__':
    main()


import numpy as np
from sklearn.preprocessing import OneHotEncoder


class OneHotEncoderEIMTC:
    def __init__(self,  **kwargs) -> None:
        self._inner_ohc = OneHotEncoder(**kwargs)

    @property
    def infrequent_categories_(self):
        return self._inner_ohc.infrequent_categories_
    
    @property
    def categories_(self):
        return self._inner_ohc.categories_[0]
    
    @property
    def drop_idx_(self):
        return self._inner_ohc.drop_idx_
    
    @property
    def n_features_in_(self):
        return self._inner_ohc.n_features_in_
    
    @property
    def feature_names_in_(self):
        return self._inner_ohc.feature_names_in_
    
    def fit(self, X, y=None):
        X = np.asarray(X).reshape(-1,1)
        self._inner_ohc.fit(X, y)
        
    def fit_transform(self, X, y=None):
        X = np.asarray(X).reshape(-1,1)
        return self._inner_ohc.fit_transform(X, y)
    
    def transform(self, X):
        X = np.asarray(X).reshape(-1,1)
        return self._inner_ohc.transform(X)
    
    def inverse_transform(self, X):
        return self._inner_ohc.inverse_transform(X).reshape(-1)
    
    def get_feature_names_out(self, input_features=None):
        if input_features is not None:
            input_features = [input_features]
        return self._inner_ohc.get_feature_names_out(input_features)


from EIMTC.stats.stats import IterableStats
from nfstream import NFPlugin
import numpy as np # for bytes distribution

class ResReqDiffTime(NFPlugin):
    ''' ResReqDiffTime | 
    Statistics of delta time between upload packets to download packets.
    For example, if there are 3 consequtive packets at time t1 going from src -> dst, and then the next packet is
    from the other direction, dst -> src, at tune t2 then the time calculation is the diff between the first upload packet
    to the first download packet hence, t2 -t1. This process repeats to get a list of such values. The plugin then extracts
    statistics on the series of time values.
    
    Output Features:
        udps.req_res_time_diff: (list) The time values of flow direction changes.
        udps.min_req_res_time_diff: (float)
        udps.max_req_res_time_diff: (float)
        udps.mean_req_res_time_diff: (float)
        udps.stddev_req_res_time_diff: (float)
        udps.variance_req_res_time_diff: (float)
        udps.coeff_of_var_req_res_time_diff: (float)
        udps.skew_from_median_req_res_time_diff: (float)
        
        Name Format:
            udps.[STAT]_req_res_time_diff
        
        Statistics (STAT) Names:
            - max
            - min
            - mean
            - stddev
            - variance
            - coeff_of_var
            - skew_from_median
    '''
    def __init__(self, **kwargs):
        ''' '''
        super().__init__(**kwargs)

    def on_init(self, packet, flow):
        ''' '''
        flow.udps.req_res_time_diff = list() 
        flow.udps.current_flow_direction = 0 # 0 for forward, 1 for backward
        flow.udps.current_flow_direction_timestamp = packet.time
        flow.udps.min_req_res_time_diff = 0
        flow.udps.max_req_res_time_diff = 0
        flow.udps.mean_req_res_time_diff = 0
        flow.udps.stddev_req_res_time_diff = 0
        flow.udps.variance_req_res_time_diff = 0
        flow.udps.coeff_of_var_req_res_time_diff = 0
        flow.udps.skew_from_median_req_res_time_diff = 0

    def on_update(self, packet, flow):
        ''' '''
        if packet.direction != flow.udps.current_flow_direction:
            flow.udps.req_res_time_diff.append(packet.time - flow.udps.current_flow_direction_timestamp)
            flow.udps.current_flow_direction = packet.direction
            flow.udps.current_flow_direction_timestamp = packet.time

    def on_expire(self, flow): 
        ''' '''
        stats = IterableStats(flow.udps.req_res_time_diff)
        flow.udps.min_req_res_time_diff = stats.min()
        flow.udps.max_req_res_time_diff = stats.max()
        flow.udps.mean_req_res_time_diff = stats.average()
        flow.udps.median_req_res_time_diff = stats.median()
        flow.udps.stddev_req_res_time_diff = stats.std_deviation()
        flow.udps.variance_req_res_time_diff = stats.variance()
        flow.udps.coeff_of_var_req_res_time_diff = stats.coeff_of_variation()
        flow.udps.skew_from_median_req_res_time_diff = stats.skew_from_median()
        
        # Cleanup
        del flow.udps.current_flow_direction_timestamp
        del flow.udps.current_flow_direction






from tensorflow.keras.models import Model
from tensorflow.keras.layers import Layer, BatchNormalization, Concatenate, Dropout, Dense, Softmax
from tensorflow.keras import activations
import tensorflow as tf


class GraphDApp(Model):
    '''
    # GraphDApp Model
    ## Parameters
    
    - `n_classes` (int): Number of possible classes an input can be classified into (#units in final dense layer). 
    - `mlplayer_units` (int): Number of units in each single-MLP layer.
    
    ## Input
    ### Per sample
    - Graph adjacancy matrix
        - Shape: (n_nodes, n_nodes)
    - Node features
        - Shape: (n_nodes, n_features_per_node)
    
    ### Final input shape
    [(n_samples, n_nodes, n, nodes), (n_samples, n_nodes, n_features_per_node)]
    
    
    ## Paper
    
    "Accurate Decentralized Application Identification via Encrypted Traffic Analysis Using Graph Neural Networks,"
    
    ### By
    
    - Meng Shen.
    - Jinpeng Zhang.
    - Liehuang Zhu.
    - Ke Xu.
    - Xiaojiang Du.
    '''
    def __init__(
            self, 
            n_classes, 
            mlplayer_units=64, 
            neighbour_agg_method='sum', 
            **kwargs):
        super(GraphDApp, self).__init__(**kwargs)
        self.mlp1 = MLPLayer(mlplayer_units)
        self.mlp2 = MLPLayer(mlplayer_units)
        self.mlp3 = MLPLayer(mlplayer_units)
        self.readout = Readout()
        self.concat = Concatenate()
        self.dense = Dense(n_classes)
        self.classification = Softmax()
    
    def call(self, inputs):
        x1 = self.mlp1(inputs)
        x2 = self.mlp2(x1)
        x3 = self.mlp3(x2)
        x4 = self.readout([x1, x2, x3])
        x4 = self.concat(x4)
        x4 = self.dense(x4)
        return self.classification(x4)


class Readout(Layer):
    def __init__(self):
        super(Readout, self).__init__()
    
    def call(self, inputs):
        FsBatch = []
        for i in range(len(inputs)):
            Fbatch = inputs[i][1]
            FsBatch.append(tf.reduce_sum(Fbatch, axis=-2))
        return FsBatch


class MLPLayer(Layer):
    def __init__(
            self, 
            output_size=5, 
            activation='relu', 
            use_bias=True, 
            neighbour_agg_method='sum', 
            dropout_rate=0.025):
        super(MLPLayer, self).__init__()
        self.output_size = output_size
        self.dense = Dense(self.output_size, use_bias=use_bias)
        self.activation = activations.get(activation)
        self.batch_norm = BatchNormalization()
        self.dropout = Dropout(dropout_rate)
    
    def build(self, input_shape):
        self.eps = self.add_weight(
            name='epsilon',
            shape=(1,),
            initializer="random_normal",
            dtype='float32',
            trainable=True,
        )
    
    def call(self, inputs, training=False):
        '''
        [A, F]
        '''
        A = inputs[0]
        F = inputs[1]
        outputs = tf.multiply(1.0+self.eps, F) + tf.matmul(A,F)
        outputs = self.dense(outputs)
        outputs = self.activation(outputs)
        outputs = self.batch_norm(outputs, training=training)
        outputs = self.dropout(outputs)
        return [A, outputs]


        
class DistillerPreprocessing:
    @staticmethod
    def preprocess_features(dataframe):
        from EIMTC.plugins.n_bytes import NBytes
        from EIMTC.plugins.protocol_header_fields import ProtocolHeaderFields
        NBytes.preprocess(dataframe)
        ProtocolHeaderFields.preprocess(dataframe)
        



from nfstream import NFStreamer, NFPlugin
import numpy as np
from scapy.all import IP, TCP

class ProtocolHeaderFields(NFPlugin):
    ''' ProtocolHeaderFields |
    Extracts 4 features for `n_packets` packets. Packet direction, payload size, delta time
    and TCP window size if TCP, value is set to zero if UDP. 
    
    Feature Outputs:
        - udps.protocol_header_fields: utput shape of (`n_packets`, 4). Hence `n_packets` rows and 4 columns per row where each column contains the respective feature extracted from the packet. [dir, size, time, win]
    
    Paper:
        "Network Traffic Classifier With Convolutional and Recurrent Neural Networks for Internet of Things"
        
        By:
            - Manuel Lopez-Martin (Senior Member, IEEE).
            - Belen Carro.
            - Antonio Sanchez-Esguevillas (Senior Member, IEEE).
            - Jaime Lloret (Senior Member, IEEE).
    '''
    def __init__(self, n_packets=32):
        '''
        Args:
            `n_packets` (int): The number of packets to process.
        '''
        self.n_packets = n_packets
    
    def on_init(self, packet, flow):
        ''' '''
        flow.udps.protocol_header_fields_n_packets = self.n_packets
        flow.udps.protocol_header_fields = np.zeros((self.n_packets,4), dtype=np.uint32)
        
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        ''' '''
        if flow.bidirectional_packets <= self.n_packets and packet.protocol in [6, 17]: # TCP or UDP only.
            flow.udps.protocol_header_fields[flow.bidirectional_packets-1] = [
                packet.direction,
                packet.payload_size,
                packet.delta_time,   
                IP(packet.ip_packet)[TCP].window if packet.protocol==6 else 0
            ]

    def on_expire(self, flow):
        ''' '''
        matrix = []
        for row in flow.udps.protocol_header_fields:
            matrix.append(list(row))
            
        flow.udps.protocol_header_fields = matrix


    @staticmethod
    def preprocess(dataframe):
        import ast
        # validate
        assert 'udps.protocol_header_fields' in dataframe.columns, "Column 'udps.protocol_header_fields' not found."
        assert isinstance(dataframe['udps.protocol_header_fields'].iloc[0], str), "Values in column 'udps.protocol_header_fields' are already processed."
        
        dataframe['udps.protocol_header_fields'] = dataframe['udps.protocol_header_fields'].apply(ast.literal_eval)
        
        